@@ -1,6 +1,7 @@
 // Copyright 2018-2021 the Deno authors. All rights reserved. MIT license.
 
 // @ts-check
+/// <reference path="../../core/internal.d.ts" />
 /// <reference path="../../core/lib.deno_core.d.ts" />
 /// <reference path="../webidl/internal.d.ts" />
 /// <reference path="../web/internal.d.ts" />
@@ -9,9 +10,15 @@
 "use strict";
 
 ((window) => {
+  const {
+    Error,
+    ObjectDefineProperty,
+    ObjectEntries,
+    ObjectSetPrototypeOf,
+  } = window.__bootstrap.primordials;
   const webidl = window.__bootstrap.webidl;
+  const consoleInternal = window.__bootstrap.console;
 
-  const { defineProperty } = Object;
   // Defined in WebIDL 4.3.
   // https://heycam.github.io/webidl/#idl-DOMException
   const INDEX_SIZE_ERR = 1;
@@ -70,13 +77,12 @@
 
   // Defined in WebIDL 4.3.
   // https://heycam.github.io/webidl/#idl-DOMException
-  class DOMException extends Error {
+  class DOMException {
     #message = "";
     #name = "";
     #code = 0;
 
     constructor(message = "", name = "Error") {
-      super();
       this.#message = webidl.converters.DOMString(message, {
         prefix: "Failed to construct 'DOMException'",
         context: "Argument 1",
@@ -103,12 +109,30 @@
     get [Symbol.toStringTag]() {
       return "DOMException";
     }
+
+    [Symbol.for("Deno.customInspect")](inspect) {
+      if (this instanceof DOMException) {
+        return `DOMException: ${this.#message}`;
+      } else {
+        return inspect(consoleInternal.createFilteredInspectProxy({
+          object: this,
+          evaluate: false,
+          keys: [
+            "message",
+            "name",
+            "code",
+          ],
+        }));
+      }
+    }
   }
 
+  ObjectSetPrototypeOf(DOMException.prototype, Error.prototype);
+
   webidl.configurePrototype(DOMException);
 
   for (
-    const [key, value] of Object.entries({
+    const [key, value] of ObjectEntries({
       INDEX_SIZE_ERR,
       DOMSTRING_SIZE_ERR,
       HIERARCHY_REQUEST_ERR,
@@ -137,10 +161,9 @@
     })
   ) {
     const desc = { value, enumerable: true };
-    defineProperty(DOMException, key, desc);
-    defineProperty(DOMException.prototype, key, desc);
+    ObjectDefineProperty(DOMException, key, desc);
+    ObjectDefineProperty(DOMException.prototype, key, desc);
   }
 
-  window.DOMException = DOMException;
-  defineProperty(window, "DOMException", { enumerable: false });
+  window.__bootstrap.domException = { DOMException };
 })(this);
